/*
 * ALSA SEQ < - > JACK MIDI bridge
 *
 * Copyright (c) 2006,2007 Dmitry S. Baikov <c0ff@konstruktiv.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

/*
 * input_thread:
 *   if event is about port add/del, post port_event to port_events, wake port_thread
 *   else post midi_event to input ring
 * port_thread:
 *   read all port_events:
 *     insert added ports to hash, and send them to add_ports ring
 *     mark deleted ports as is_dead
 *   all ports from del_ports rings:
 *     remove from hash
 *     delete
 * jack_input:
 *   move all is_dead ports to del_ports ring
 *   read and dispatch all midi_events (reads port hash)
 * jack_output:
 *   sort and send all midi_events to output ring
 *   wake output thread
 * output_thread:
 *   send all midi_events
 *   drain output
 *   wait
 */
#include <alsa/asoundlib.h>
#include <jack/jack.h>
#include <jack/midiport.h>
#include <jack/ringbuffer.h>
#include <jack/thread.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <semaphore.h>
#include <time.h>

#include "./alsa_midi.h"

//#define USE_SELECT
//#define USE_POLL
//#define USE_ALSA_QUEUE

//#define USE_JACK_TIME

#if defined(USE_ALSA_QUEUE) && !defined(USE_JACK_TIME)
#warning USE_ALSA_QUEUE without USE_JACK_TIME can lead to event reordering.
#endif

enum {
	MAX_PORTS = 64,
	MAX_EVENTS = 256,
	MAX_EVENT_SIZE = 1024,
};

typedef struct port_t port_t;

enum {
	PORT_HASH_BITS = 4,
	PORT_HASH_SIZE = 1 << PORT_HASH_BITS
};

typedef port_t* port_hash_t[PORT_HASH_SIZE];


struct port_t {
	port_t *next;
	int is_dead;
	char name[64];
	snd_seq_addr_t remote;
	int alsa_port;
	jack_port_t *jack_port;

	void *jack_buf;

	/* input */
	//int lost_events;

	/* output */
	int nevents;
	int index;
	jack_midi_event_t event;
};

enum PortAction { PORT_ADD, PORT_DEL };

typedef struct {
	enum PortAction action;
	snd_seq_addr_t addr;
} port_event_t;

typedef struct {
	jack_time_t time;
	int size;
	snd_seq_addr_t addr;
} midi_event_t;

typedef struct {
	pthread_t thread;

	jack_ringbuffer_t *add_ports;
	jack_ringbuffer_t *del_ports;

	jack_ringbuffer_t *event_ring;
	jack_ringbuffer_t *data_ring;

	port_hash_t midi_ports;
	int jack_nports;
	port_t *jack_ports[MAX_PORTS];
} stream_t;

enum { PORT_THREAD=1, INPUT_THREAD=2, OUTPUT_THREAD=4 };

typedef struct alsa_seqmidi {
	alsa_midi_t ops;
	jack_client_t *jack;

	snd_seq_t *seq;
	int client_id;
	int announce;
#ifdef USE_ALSA_QUEUE
	int input_que;
	int output_que;
#endif

	int keep_walking;
	int threads_run;

	pthread_t port_thread;
	sem_t port_sem;
	jack_ringbuffer_t *port_events;

	sem_t out_sem;

	stream_t stream[2];
} alsa_seqmidi_t;


enum PortType { PORT_INPUT = 0, PORT_OUTPUT = 1 };

typedef void (*port_jack_func)(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes);
static void do_jack_input(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes);
static void do_jack_output(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes);

typedef struct {
	int alsa_mask;
	int alsa_caps;
	int jack_caps;
	char name[4];
	port_jack_func jack_func;
} port_type_t;

static port_type_t port_type[2] = {
	{
		SND_SEQ_PORT_CAP_SUBS_READ,
		SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
		JackPortIsOutput|JackPortIsPhysical|JackPortIsTerminal,
		"in",
		do_jack_input
	},
	{
		SND_SEQ_PORT_CAP_SUBS_WRITE,
		SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
		JackPortIsInput|JackPortIsPhysical|JackPortIsTerminal,
		"out",
		do_jack_output
	}
};


static void alsa_seqmidi_delete(alsa_midi_t *m);
static int alsa_seqmidi_attach(alsa_midi_t *m);
static int alsa_seqmidi_detach(alsa_midi_t *m);
static int alsa_seqmidi_start(alsa_midi_t *m);
static int alsa_seqmidi_stop(alsa_midi_t *m);
static void alsa_seqmidi_read(alsa_midi_t *m, jack_nframes_t nframes);
static void alsa_seqmidi_write(alsa_midi_t *m, jack_nframes_t nframes);

static
void stream_init(alsa_seqmidi_t *self, int dir)
{
	stream_t *str = &self->stream[dir];

	str->add_ports = jack_ringbuffer_create(MAX_PORTS*sizeof(port_t*));
	str->del_ports = jack_ringbuffer_create(MAX_PORTS*sizeof(port_t*));

	str->event_ring = jack_ringbuffer_create(MAX_EVENTS*sizeof(midi_event_t));
	str->data_ring = jack_ringbuffer_create(MAX_EVENT_SIZE*32);
}

static
void stream_stop(alsa_seqmidi_t *self, int dir)
{
	stream_t *str = &self->stream[dir];

	jack_ringbuffer_reset(str->event_ring);
	jack_ringbuffer_reset(str->data_ring);
}

static void port_free(alsa_seqmidi_t *self, port_t *port);
static void free_ports(alsa_seqmidi_t *self, int type);

static
void stream_detach(alsa_seqmidi_t *self, int dir)
{
	stream_t *str = &self->stream[dir];
	int i;

	// delete all ports in del_ports queue
	free_ports(self, dir);

	// delete all is_dead ports from jack_ports (they are not in hash already)
	for (i=0; i<str->jack_nports; ++i) {
		port_t *port = str->jack_ports[i];
		if (port->is_dead)
			port_free(self, port);
	}
	str->jack_nports = 0;

	// new ports are in hash
	jack_ringbuffer_reset(str->add_ports);

	// delete all ports from hash
	for (i=0; i<PORT_HASH_SIZE; ++i) {
		port_t *port = str->midi_ports[i];
		while (port) {
			port_t *next = port->next;
			port_free(self, port);
			port = next;
		}
		str->midi_ports[i] = NULL;
	}
}

static
void stream_close(alsa_seqmidi_t *self, int dir)
{
	stream_t *str = &self->stream[dir];

	if (str->data_ring)
		jack_ringbuffer_free(str->data_ring);
	if (str->event_ring)
		jack_ringbuffer_free(str->event_ring);
	if (str->add_ports)
		jack_ringbuffer_free(str->add_ports);
	if (str->del_ports)
		jack_ringbuffer_free(str->del_ports);
}

alsa_midi_t* alsa_seqmidi_new(jack_client_t *client)
{
	alsa_seqmidi_t *self = calloc(1, sizeof(alsa_seqmidi_t));
	if (!self)
		return NULL;
	self->jack = client;

	self->port_events = jack_ringbuffer_create(2*MAX_PORTS*sizeof(port_t*));

	stream_init(self, PORT_INPUT);
	stream_init(self, PORT_OUTPUT);

	sem_init(&self->port_sem, 0, 0);
	sem_init(&self->out_sem, 0, 0);

	self->ops.destroy = alsa_seqmidi_delete;
	self->ops.attach = alsa_seqmidi_attach;
	self->ops.detach = alsa_seqmidi_detach;
	self->ops.start = alsa_seqmidi_start;
	self->ops.stop = alsa_seqmidi_stop;
	self->ops.read = alsa_seqmidi_read;
	self->ops.write = alsa_seqmidi_write;
	return &self->ops;
}

static
void alsa_seqmidi_delete(alsa_midi_t *m)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;

	alsa_seqmidi_detach(m);

	stream_close(self, PORT_OUTPUT);
	stream_close(self, PORT_INPUT);

	jack_ringbuffer_free(self->port_events);

	sem_close(&self->out_sem);
	sem_close(&self->port_sem);

	free(self);
}

static void add_existing_ports(alsa_seqmidi_t *self);

static
int alsa_seqmidi_attach(alsa_midi_t *m)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;
	int err;

	if (self->seq)
		return -EALREADY;

	if ((err = snd_seq_open(&self->seq, "hw", SND_SEQ_OPEN_DUPLEX, 0)) < 0) {
		printf("failed to open alsa seq");
		return err;
	}
	snd_seq_set_client_name(self->seq, "jack-alsaseq");
	self->announce = snd_seq_create_simple_port(self->seq, "announce",
		SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|SND_SEQ_PORT_CAP_NO_EXPORT,
		SND_SEQ_PORT_TYPE_APPLICATION);
	self->client_id = snd_seq_client_id(self->seq);

#ifdef USE_ALSA_QUEUE
  	self->input_que = snd_seq_alloc_queue(self->seq);
  	snd_seq_start_queue(self->seq, self->input_que, 0); 

  	self->output_que = snd_seq_alloc_queue(self->seq);
  	snd_seq_start_queue(self->seq, self->output_que, 0);
#endif

	snd_seq_nonblock(self->seq, 1);

	add_existing_ports(self);

	return 0;
}

static
int alsa_seqmidi_detach(alsa_midi_t *m)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;

	if (!self->seq)
		return -EALREADY;

	alsa_seqmidi_stop(m);

	stream_detach(self, PORT_INPUT);
	stream_detach(self, PORT_OUTPUT);
	jack_ringbuffer_reset(self->port_events);

	snd_seq_close(self->seq);
	self->seq = NULL;

	return 0;
}

static void* port_thread(void *);
static void* input_thread(void *);
static void* output_thread(void *);

static
int alsa_seqmidi_start(alsa_midi_t *m)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;
	int err;
	int prio = 99;
	int rt = jack_is_realtime(self->jack);

	if (!self->seq)
		return -EBADF;

	if (self->keep_walking)
		return -EALREADY;

	assert (self->threads_run == 0);

	snd_seq_drop_input(self->seq);

	self->keep_walking = 1;

	if ((err = jack_client_create_thread(self->jack, &self->port_thread, 0, 0, port_thread, self)))
		goto fail;
	self->threads_run |= PORT_THREAD;

	if ((err = jack_client_create_thread(self->jack, &self->stream[PORT_INPUT].thread, prio, rt, input_thread, self)))
		goto fail;
	self->threads_run |= INPUT_THREAD;
	snd_seq_connect_from(self->seq, self->announce, SND_SEQ_CLIENT_SYSTEM, SND_SEQ_PORT_SYSTEM_ANNOUNCE);
	
	if ((err = jack_client_create_thread(self->jack, &self->stream[PORT_OUTPUT].thread,prio, rt, output_thread, self)))
		goto fail;
	self->threads_run |= OUTPUT_THREAD;
	
	return 0;

 fail:
 	alsa_seqmidi_stop(m);
 	return -err;
}

static
int alsa_seqmidi_stop(alsa_midi_t *m)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;
	if (!self->keep_walking)
		return -EALREADY;

	self->keep_walking = 0;

	if (self->threads_run & OUTPUT_THREAD) {
		sem_post(&self->out_sem);
		pthread_join(self->stream[PORT_OUTPUT].thread, NULL);
		self->threads_run &= ~OUTPUT_THREAD;
	}

	if (self->threads_run & INPUT_THREAD) {
		snd_seq_disconnect_from(self->seq, self->announce,
			SND_SEQ_CLIENT_SYSTEM, SND_SEQ_PORT_SYSTEM_ANNOUNCE);
		pthread_join(self->stream[PORT_INPUT].thread, NULL);
		self->threads_run &= ~INPUT_THREAD;
	}

	if (self->threads_run & PORT_THREAD) {
		sem_post(&self->port_sem);
		pthread_join(self->port_thread, NULL);
		self->threads_run &= ~PORT_THREAD;
	}

	stream_stop(self, PORT_INPUT);
	stream_stop(self, PORT_OUTPUT);

	assert (self->threads_run == 0);
	return 0;
}


#ifdef USE_ALSA_QUEUE
static void alsa_connect(snd_seq_t *seq, int que, int c1, int p1, int c2, int p2)
{
	snd_seq_port_subscribe_t* sub;
	snd_seq_addr_t seq_addr;
	int err;

	snd_seq_port_subscribe_alloca(&sub);
	seq_addr.client = c1;
	seq_addr.port = p1;
	snd_seq_port_subscribe_set_sender(sub, &seq_addr);
	seq_addr.client = c2;
	seq_addr.port = p2;
	snd_seq_port_subscribe_set_dest(sub, &seq_addr);
	if (que >= 0)
		snd_seq_port_subscribe_set_queue(sub, que);
	if ((err=snd_seq_subscribe_port(seq, sub)))
		printf("can't subscribe %d:%d to %d:%d - %s\n", c1, p1, c2, p2, snd_strerror(err));
}
#endif

/*
 * ==================== Port routines =============================
 */
static inline
int port_hash(snd_seq_addr_t addr)
{
	return (addr.client + addr.port) % PORT_HASH_SIZE;
}

static
port_t** port_find(port_hash_t hash, snd_seq_addr_t addr)
{
	port_t **pport = &hash[port_hash(addr)];
	while (*pport) {
		port_t *port = *pport;
		if ( port->remote.client == addr.client && port->remote.port == addr.port)
			return pport;
		pport = &port->next;
	}
	return NULL;
}

static
void port_insert(port_hash_t hash, port_t *port)
{
	port_t **pport = &hash[port_hash(port->remote)];
	port->next = *pport;
	*pport = port;
}

static
port_t* port_remove(port_hash_t hash, snd_seq_addr_t addr)
{
	port_t **pport = port_find(hash, addr);
	if (pport) {
		port_t *port = *pport;
		*pport = port->next;
		return port;
	}
	return NULL;
}

static
void port_free(alsa_seqmidi_t *self, port_t *port)
{
	if (port->jack_port)
		jack_port_unregister(self->jack, port->jack_port);
	if (port->alsa_port >= 0)
		snd_seq_delete_simple_port(self->seq, port->alsa_port);
	printf("port deleted: %s\n", port->name);

	free(port);
}

static
port_t* port_create(alsa_seqmidi_t *self, int type, snd_seq_addr_t addr, const snd_seq_port_info_t *info)
{
	port_t *port = calloc(1, sizeof(port_t));
	if (!port)
		return NULL;

	port->alsa_port = -1;
	port->remote = addr;

	snprintf(port->name, sizeof(port->name), "%s (%d:%d) %s",
		port_type[type].name, addr.client, addr.port, snd_seq_port_info_get_name(info));

	port->alsa_port = snd_seq_create_simple_port(
		self->seq, port->name,
#ifdef DEBUG
			port_type[type].alsa_caps,
#else
			port_type[type].alsa_caps | SND_SEQ_PORT_CAP_NO_EXPORT,
#endif
			SND_SEQ_PORT_TYPE_APPLICATION);

	if (port->alsa_port < 0)
		goto failed;

	port->jack_port = jack_port_register(self->jack,
		port->name, JACK_DEFAULT_MIDI_TYPE, port_type[type].jack_caps, 0);
	if (!port->jack_port)
		goto failed;

#ifdef USE_ALSA_QUEUE
	if (type == PORT_INPUT)
		alsa_connect (self->seq, self->input_que,
			port->remote.client, port->remote.port, self->client_id, port->alsa_port);
	else
		alsa_connect (self->seq, self->output_que,
			self->client_id, port->alsa_port, port->remote.client, port->remote.port);
#else
	if (type == PORT_INPUT)
		snd_seq_connect_from(self->seq, port->alsa_port, port->remote.client, port->remote.port);
	else
		snd_seq_connect_to(self->seq, port->alsa_port, port->remote.client, port->remote.port);
#endif
	printf("port created: %s\n", port->name);
	return port;

 failed:
 	port_free(self, port);
	return NULL;
}

/*
 * ==================== Port add/del handling thread ==============================
 */
static
void remove_port(alsa_seqmidi_t *self, int type, snd_seq_addr_t addr)
{
	port_t *port = port_remove(self->stream[type].midi_ports, addr);
	if (port)
		port->is_dead = 1; // see jack_process
}

static
void update_port_type(alsa_seqmidi_t *self, int type, snd_seq_addr_t addr, int caps, const snd_seq_port_info_t *info)
{
	port_t **pport = port_find(self->stream[type].midi_ports, addr);
	int alsa_mask = port_type[type].alsa_mask;
	if (pport && (caps & alsa_mask)!=alsa_mask) {
		remove_port(self, type, addr);
	}
	if (!pport && (caps & alsa_mask)==alsa_mask) {
		port_t *port = port_create(self, type, addr, info);
		if (port) {
			port_insert(self->stream[type].midi_ports, port);
			assert (jack_ringbuffer_write_space(self->stream[type].add_ports) >= sizeof(port));
			jack_ringbuffer_write(self->stream[type].add_ports, (char*)&port, sizeof(port));
		}
	}
}

static
void update_port(alsa_seqmidi_t *self, snd_seq_addr_t addr, const snd_seq_port_info_t *info)
{
	unsigned int port_caps = snd_seq_port_info_get_capability(info);
	if (port_caps & SND_SEQ_PORT_CAP_NO_EXPORT)
		return;
	update_port_type(self, PORT_INPUT, addr, port_caps, info);
	update_port_type(self, PORT_OUTPUT, addr, port_caps, info);
}

static
void free_ports(alsa_seqmidi_t *self, int type)
{
	jack_ringbuffer_t *ring = self->stream[type].del_ports;
	port_t *port;
	int sz;
	while ((sz = jack_ringbuffer_read(ring, (char*)&port, sizeof(port)))) {
		assert (sz == sizeof(port));
		port_free(self, port);
	}
}

static
void* port_thread(void *arg)
{
	alsa_seqmidi_t *self = arg;

	add_existing_ports(self);

	while (self->keep_walking) {
		port_event_t ev;
		int sz;

		sem_wait(&self->port_sem);
		
		while ((sz = jack_ringbuffer_read(self->port_events, (char*)&ev, sizeof(ev)))) {
			assert (sz == sizeof(ev));
			if (ev.action == PORT_ADD) {
				snd_seq_port_info_t *info;
				snd_seq_port_info_alloca(&info);
				int err;
				if ((err=snd_seq_get_any_port_info(self->seq, ev.addr.client, ev.addr.port, info))>=0)
					update_port(self, ev.addr, info);
			} else {
				remove_port(self, PORT_INPUT, ev.addr);
				remove_port(self, PORT_OUTPUT, ev.addr);
			}
		}

		free_ports(self, PORT_INPUT);
		free_ports(self, PORT_OUTPUT);
	}
#ifdef DEBUG
	printf("port_thread exited\n");
#endif
	return NULL;
}

static
void add_existing_ports(alsa_seqmidi_t *self)
{
	snd_seq_addr_t addr;
	snd_seq_client_info_t *client_info;
	snd_seq_port_info_t *port_info;

	snd_seq_client_info_alloca(&client_info);
	snd_seq_port_info_alloca(&port_info);
	snd_seq_client_info_set_client(client_info, -1);
	while (snd_seq_query_next_client(self->seq, client_info) >= 0)
	{
		addr.client = snd_seq_client_info_get_client(client_info);
		if (addr.client == SND_SEQ_CLIENT_SYSTEM || addr.client == self->client_id)
			continue;
		snd_seq_port_info_set_client(port_info, addr.client);
		snd_seq_port_info_set_port(port_info, -1);
		while (snd_seq_query_next_port(self->seq, port_info) >= 0)
		{
			addr.port = snd_seq_port_info_get_port(port_info);
			update_port(self, addr, port_info);
		}
	}
}

/*
 * =================== Input/output port handling =========================
 */
static
port_t* port_get(port_hash_t hash, snd_seq_addr_t addr)
{
	port_t **pport = &hash[port_hash(addr)];
	while (*pport) {
		port_t *port = *pport;
		if (port->remote.client == addr.client && port->remote.port == addr.port)
			return port;
		pport = &port->next;
	}
	return NULL;
}

static
int jack_process(alsa_seqmidi_t *self, int dir, jack_nframes_t nframes)
{
	stream_t *str = &self->stream[dir];
	port_jack_func process = port_type[dir].jack_func;
	int r, w, res, del=0;
	port_t *port;

	// process ports
	for (r=0, w=0; r<str->jack_nports; ++r) {
		port = str->jack_ports[r];
		port->jack_buf = jack_port_get_buffer(port->jack_port, nframes);

		if (dir == PORT_INPUT)
			jack_midi_clear_buffer(port->jack_buf, nframes);

		if (port->is_dead) {
			jack_ringbuffer_t *ring = str->del_ports;
			assert (jack_ringbuffer_write_space(ring) >= sizeof(port));
#ifdef DEBUG
			//printf("jack: removed port %s\n", port->name);
#endif
			jack_ringbuffer_write(ring, (char*)&port, sizeof(port));
			++del;
			continue;
		}

		if (r != w)
			str->jack_ports[w] = port;
		++w;

		(*process)(self, port, nframes);
	}
	res = w;

	if (del)
		sem_post(&self->port_sem);

	// add new ports
	while (jack_ringbuffer_read(str->add_ports, (char*)&port, sizeof(port)) && w < MAX_PORTS) {
		str->jack_ports[w++] = port;
	}
	str->jack_nports = w;

	return res;
}

static
int poll_seq(alsa_seqmidi_t *self, struct pollfd *pfds, int npfds, int timeout)
{
	int err;
	unsigned short revents = 0;

	do {
		err = poll(pfds, npfds, timeout);
		if (!self->keep_walking)
			return -1;
		if (err < 0) {
			if (errno == EINTR)
				continue;
			perror("poll");
			return -errno;
		}
	} while (err <= 0);

	snd_seq_poll_descriptors_revents(self->seq, pfds, npfds, &revents);
	return revents;
}

/*
 * ============================ Input ==============================
 */
static
void do_jack_input(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes)
{
	//port->lost_events = 0;
}

static
void alsa_seqmidi_read(alsa_midi_t *m, jack_nframes_t nframes)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;
	stream_t *str = &self->stream[PORT_INPUT];
	midi_event_t event;
	jack_nframes_t cur_frame_time = jack_last_frame_time(self->jack);

	if (!self->keep_walking)
		return;

	jack_process((alsa_seqmidi_t*)m, PORT_INPUT, nframes); 

	while (jack_ringbuffer_read(str->event_ring, (char*)&event, sizeof(event))) {
		jack_midi_data_t *buf;
		jack_nframes_t time;

		port_t *port = port_get(str->midi_ports, event.addr);
		if (!port) {
			jack_ringbuffer_read_advance(str->data_ring, event.size);
			continue;
		}
		
		//port->lost_events += event.lost_events;

		if ((event.time + nframes) < cur_frame_time)
			time = 0;
		else if (event.time >= cur_frame_time)
			time = nframes - 1;
		else
			time = event.time + nframes - cur_frame_time;

		buf = jack_midi_event_reserve(port->jack_buf, time, event.size, nframes);
		if (buf)
			jack_ringbuffer_read(str->data_ring, (char*)buf, event.size);
		else {
			jack_ringbuffer_read_advance(str->data_ring, event.size);	
			//port->lost_events++;
		}
#ifdef DEBUG
		//printf("jack event: %d bytes at %d\n", (int)event.size, (int)time);
#endif
	}
	//jack_midi_set_lost_events(port_buf, port->lost_events, nframes);
}

static
void input_event(alsa_seqmidi_t *self, snd_seq_event_t *ev, snd_midi_event_t *decoder)
{
	midi_event_t event;
	jack_midi_data_t data[MAX_EVENT_SIZE];
	stream_t *str = &self->stream[PORT_INPUT];
	long size;
	/*
	 * RPNs, NRPNs, Bank Change need special handling
	 * but seems, ALSA does it for us already.
	 */
	snd_midi_event_reset_decode(decoder);
	if ((size = snd_midi_event_decode(decoder, data, sizeof(data), ev))<0)
		return;

	// fixup NoteOn with vel 0
	if (data[0] == 0x90 && data[2] == 0x00) {
		data[0] = 0x80;
		data[2] = 0x40;
	}
	if (jack_ringbuffer_write_space(str->event_ring) < sizeof(event)
	  || jack_ringbuffer_write_space(str->data_ring) < size) {
	  	//str->lost_events++;
	  	return;
	}
	event.time = jack_frame_time(self->jack);
	event.addr = ev->source;
	event.size = size;
	//event.lost_events = str->lost_events;
	//str->lost_events = 0;
	jack_ringbuffer_write(str->data_ring, (char*)data, size);
	jack_ringbuffer_write(str->event_ring, (char*)&event, sizeof(event));
#ifdef DEBUG
	//printf("midi event: %d bytes\n", (int)size);
#endif
}

static
void port_event(alsa_seqmidi_t *self, snd_seq_event_t *ev)
{
	port_event_t cmd;

	switch (ev->type) {
	case SND_SEQ_EVENT_PORT_START:
	case SND_SEQ_EVENT_PORT_CHANGE:
		cmd.action = PORT_ADD;
		break;
	case SND_SEQ_EVENT_PORT_EXIT:
		cmd.action = PORT_DEL;
		break;
	default:
		return;
	}
	if (ev->data.addr.client == self->client_id)
		return;

	cmd.addr = ev->data.addr;
#ifdef DEBUG
	printf("port_event: %s %d:%d\n", cmd.action==PORT_ADD ? "ADD":"DEL", cmd.addr.client, cmd.addr.port);
#endif
	assert (jack_ringbuffer_write_space(self->port_events) >= sizeof(cmd));
	jack_ringbuffer_write(self->port_events, (char*)&cmd, sizeof(cmd));

	sem_post(&self->port_sem);
}

static
void* input_thread(void *arg)
{
	alsa_seqmidi_t *self = arg;
	snd_midi_event_t *decoder;
	struct pollfd pfd;

	snd_seq_poll_descriptors(self->seq, &pfd, 1, POLLIN);
	snd_midi_event_new(MAX_EVENT_SIZE, &decoder);

	while (self->keep_walking) {
		snd_seq_event_t *ev;
		int res;

		res = poll_seq(self, &pfd, 1, 500);
		if (res != POLLIN)
			break;

		while ((res = snd_seq_event_input(self->seq, &ev))>0) {
			if (ev->source.client == SND_SEQ_CLIENT_SYSTEM)
				port_event(self, ev);
			else
				input_event(self, ev, decoder);
		}
	}

	snd_midi_event_free(decoder);
#ifdef DEBUG
	printf("input thread exited\n");
#endif
	return NULL;
}

/*
 * ============================ Output ==============================
 */
static
void do_jack_output(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes)
{
	port->nevents = jack_midi_get_event_count(port->jack_buf, nframes);
	port->index = 0;
	if (port->nevents)
		jack_midi_event_get(&port->event, port->jack_buf, 0, nframes);
}

static
void alsa_seqmidi_write(alsa_midi_t *m, jack_nframes_t nframes)
{
	alsa_seqmidi_t *self = (alsa_seqmidi_t*) m;
	stream_t *str = &self->stream[PORT_OUTPUT];
	jack_nframes_t cur_frame_time = jack_last_frame_time(self->jack);
	int nports;

	if (!self->keep_walking)
		return;

	nports = jack_process(self, PORT_OUTPUT, nframes);

	while (1) {
		port_t *earliest = NULL;
		midi_event_t out;
		int i;
		for (i=0; i<nports; ++i) {
			port_t *port = str->jack_ports[i];
			if (port->index < port->nevents && (!earliest || port->event.time < earliest->event.time)) {
				earliest = port;
				continue;
			}
		}
		if (!earliest)
			break;

#ifdef USE_JACK_TIME
		out.time = jack_frames_to_time(earliest->event.time + cur_frame_time + nframes);
#else
		out.time = earliest->event.time + cur_frame_time + nframes;
#endif
		out.addr.client = self->client_id;
		out.addr.port = earliest->alsa_port;
		out.size = earliest->event.size;

	//	printf("jack out: %d at %d, to %d:%d\n", earliest->event.size, (int)earliest->event.time,
	//		earliest->remote.client, earliest->remote.port);
	
		if (jack_ringbuffer_write_space(str->data_ring) < out.size ||
		  jack_ringbuffer_write_space(str->event_ring) < sizeof(out))
			break;
		jack_ringbuffer_write(str->data_ring, (char*)earliest->event.buffer, out.size);
		jack_ringbuffer_write(str->event_ring, (char*)&out, sizeof(out));
		earliest->index++;
		if (earliest->index < earliest->nevents)
			jack_midi_event_get(&earliest->event, earliest->jack_buf, earliest->index, nframes);
		else
			earliest->event.buffer = NULL;
		earliest = NULL;
	}

	sem_post(&self->out_sem);
}


static inline
void do_output_event(alsa_seqmidi_t *self, snd_seq_event_t *alsa_event, int64_t offset)
{
#if defined(USE_POLL)
	snd_seq_ev_set_direct(alsa_event);
	offset /= 1000000;
	if (offset > 0 && offset < 1000)
		poll(NULL, 0, offset);
#elif defined(USE_SELECT)
	struct timeval delay;
	snd_seq_ev_set_direct(alsa_event);
	offset /= 1000;
	if (offset > 0 && offset < 1000000) {
		delay.tv_sec = (long) (offset / 1000000);
		delay.tv_usec = (long) (offset % 1000000);
		select(0, NULL, NULL, NULL, &delay);
	}
#elif defined(USE_ALSA_QUEUE)
	snd_seq_real_time_t delay;
	delay.tv_sec = (long) (offset / 1000000000);
	delay.tv_nsec = (long) (offset % 1000000000);
	snd_seq_ev_schedule_real(alsa_event, self->output_que, 1, &delay);
#else
	struct timespec delay;
	delay.tv_sec = (long) (offset / 1000000000);
	delay.tv_nsec = (long) (offset % 1000000000);
	snd_seq_ev_set_direct(alsa_event);
	clock_nanosleep(CLOCK_MONOTONIC, 0, &delay, NULL);
#endif
}

static
void output_events(alsa_seqmidi_t *self, snd_midi_event_t *encoder)
{
	stream_t *str = &self->stream[PORT_OUTPUT];
	jack_midi_data_t data[MAX_EVENT_SIZE];
	snd_seq_event_t alsa_event;
#ifndef USE_JACK_TIME
	long rate = jack_get_sample_rate(self->jack);
#endif
	midi_event_t event;

	while (jack_ringbuffer_read(str->event_ring, (char*)&event, sizeof(event))) {
		int err;
		int64_t cur_time;
		int64_t event_offset;
		int64_t offset;

		jack_ringbuffer_read(str->data_ring, (char*)data, event.size);

		snd_seq_ev_clear(&alsa_event);
		snd_midi_event_reset_encode(encoder);
		if (!snd_midi_event_encode(encoder, data, event.size, &alsa_event))
			continue; // EBADF?

		snd_seq_ev_set_source(&alsa_event, event.addr.port);
		snd_seq_ev_set_subs(&alsa_event);

#ifdef USE_JACK_TIME
		cur_time = jack_get_time(self->jack);
#else
		cur_time = jack_frame_time(self->jack);
#endif
		event_offset = event.time - cur_time;

		if (event_offset < 0)
			offset = 0;
		else
#ifdef USE_JACK_TIME
			offset = event_offset * 1000;
#else
			offset = (event_offset * 1000000000) / rate;
#endif

		do_output_event(self, &alsa_event, offset);

#ifdef DEBUG
		{int i; for (i=0; i<event.size; ++i) printf("%02x ", (int)data[i]);
		printf(" at %ld (%ld + %ld) (+%d)\n", (long)event.time, (long)cur_time, (long)event_offset, (int)offset);}
#endif
		err = snd_seq_event_output(self->seq, &alsa_event);
		snd_seq_drain_output(self->seq);
#ifdef DEBUG
		if (err < 0) {
			printf("midiout error - %s\n", snd_strerror(err));
		}
#endif
	}
}

static
void* output_thread(void* arg)
{
	alsa_seqmidi_t *self = arg;
	snd_midi_event_t *encoder;

	snd_midi_event_new(MAX_EVENT_SIZE, &encoder);

	while (self->keep_walking) {
		sem_wait(&self->out_sem);
		output_events(self, encoder);
		//snd_seq_drain_output(self->seq);
	}

	snd_midi_event_free(encoder);
#ifdef DEBUG
	printf("output thread exited\n");
#endif
	return NULL;
}
